package com.sun.xml.internal.ws.util.pipe;

import java.io.InputStream;
import java.net.URI;
import org.w3c.dom.ls.LSInput;
import java.net.MalformedURLException;
import com.sun.xml.internal.ws.server.SDDocumentImpl;
import javax.xml.namespace.QName;
import com.sun.xml.internal.ws.api.server.SDDocumentSource;
import java.net.URL;
import com.sun.xml.internal.stream.buffer.XMLStreamBufferResult;
import org.w3c.dom.ls.LSResourceResolver;
import com.sun.xml.internal.ws.wsdl.SDDocumentResolver;
import com.sun.istack.internal.NotNull;
import org.xml.sax.SAXException;
import com.sun.xml.internal.ws.api.message.Message;
import com.sun.xml.internal.ws.developer.ValidationErrorHandler;
import com.sun.xml.internal.ws.api.message.Packet;
import java.io.StringReader;
import java.io.Reader;
import java.util.Collection;
import java.util.Enumeration;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Attr;
import org.w3c.dom.Node;
import com.sun.istack.internal.Nullable;
import org.w3c.dom.NodeList;
import org.xml.sax.helpers.NamespaceSupport;
import org.w3c.dom.Element;
import java.util.Iterator;
import javax.xml.transform.dom.DOMSource;
import java.util.logging.Level;
import java.util.HashMap;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.stream.StreamSource;
import java.io.IOException;
import javax.xml.ws.WebServiceException;
import com.sun.xml.internal.ws.util.ByteArrayBuffer;
import org.w3c.dom.Document;
import com.sun.xml.internal.ws.api.server.SDDocument;
import javax.xml.validation.Validator;
import com.sun.xml.internal.ws.api.pipe.TubeCloner;
import com.sun.xml.internal.ws.util.xml.XmlUtil;
import com.sun.xml.internal.ws.api.pipe.Tube;
import javax.xml.validation.SchemaFactory;
import com.sun.xml.internal.ws.api.server.DocumentAddressResolver;
import com.sun.xml.internal.ws.developer.SchemaValidationFeature;
import com.sun.xml.internal.ws.api.WSBinding;
import java.util.logging.Logger;
import com.sun.xml.internal.ws.api.pipe.helper.AbstractFilterTubeImpl;

public abstract class AbstractSchemaValidationTube extends AbstractFilterTubeImpl
{
    private static final Logger LOGGER;
    protected final WSBinding binding;
    protected final SchemaValidationFeature feature;
    protected final DocumentAddressResolver resolver;
    protected final SchemaFactory sf;
    
    public AbstractSchemaValidationTube(WSBinding binding, Tube next) {
        super(next);
        this.resolver = new AbstractSchemaValidationTube.ValidationDocumentAddressResolver();
        this.binding = binding;
        this.feature = (SchemaValidationFeature)binding.getFeature(SchemaValidationFeature.class);
        this.sf = XmlUtil.allowExternalAccess(SchemaFactory.newInstance("http://www.w3.org/2001/XMLSchema"), "file", false);
    }
    
    protected AbstractSchemaValidationTube(AbstractSchemaValidationTube that, TubeCloner cloner) {
        super(that, cloner);
        this.resolver = new AbstractSchemaValidationTube.ValidationDocumentAddressResolver();
        this.binding = that.binding;
        this.feature = that.feature;
        this.sf = that.sf;
    }
    
    protected abstract Validator getValidator();
    
    protected abstract boolean isNoValidation();
    
    private Document createDOM(SDDocument doc) {
        ByteArrayBuffer bab = new ByteArrayBuffer();
        try {
            doc.writeTo(null, this.resolver, bab);
        }
        catch (IOException ioe) {
            throw new WebServiceException(ioe);
        }
        Transformer trans = XmlUtil.newTransformer();
        Source source = new StreamSource(bab.newInputStream(), null);
        DOMResult result = new DOMResult();
        try {
            trans.transform(source, result);
        }
        catch (TransformerException te) {
            throw new WebServiceException(te);
        }
        return (Document)result.getNode();
    }
    
    private void updateMultiSchemaForTns(String tns, String systemId, Map<String, List<String>> schemas) {
        List<String> docIdList = (List)schemas.get(tns);
        if (docIdList == null) {
            docIdList = new ArrayList();
            schemas.put(tns, docIdList);
        }
        docIdList.add(systemId);
    }
    
    protected Source[] getSchemaSources(Iterable<SDDocument> docs, AbstractSchemaValidationTube.MetadataResolverImpl mdresolver) {
        Map<String, DOMSource> inlinedSchemas = new HashMap();
        Map<String, List<String>> multiSchemaForTns = new HashMap();
        for (SDDocument sdoc : docs) {
            if (sdoc.isWSDL()) {
                Document dom = this.createDOM(sdoc);
                this.addSchemaFragmentSource(dom, sdoc.getURL().toExternalForm(), inlinedSchemas);
            }
            else if (sdoc.isSchema()) {
                this.updateMultiSchemaForTns(((SDDocument.Schema)sdoc).getTargetNamespace(), sdoc.getURL().toExternalForm(), multiSchemaForTns);
            }
            else {
                continue;
            }
        }
        if (AbstractSchemaValidationTube.LOGGER.isLoggable(Level.FINE)) {
            AbstractSchemaValidationTube.LOGGER.log(Level.FINE, "WSDL inlined schema fragment documents(these are used to create a pseudo schema) = {0}", inlinedSchemas.keySet());
        }
        for (DOMSource src : inlinedSchemas.values()) {
            String tns = this.getTargetNamespace(src);
            this.updateMultiSchemaForTns(tns, src.getSystemId(), multiSchemaForTns);
        }
        if (multiSchemaForTns.isEmpty()) {
            return new Source[0];
        }
        if (multiSchemaForTns.size() == 1 && ((List)multiSchemaForTns.values().iterator().next()).size() == 1) {
            String systemId2 = (String)((List)multiSchemaForTns.values().iterator().next()).get(0);
            return new Source[] { (Source)inlinedSchemas.get(systemId2) };
        }
        mdresolver.addSchemas(inlinedSchemas.values());
        Map<String, String> oneSchemaForTns = new HashMap();
        int i = 0;
        for (Map.Entry<String, List<String>> e : multiSchemaForTns.entrySet()) {
            List<String> sameTnsSchemas = (List)e.getValue();
            String systemId;
            if (sameTnsSchemas.size() > 1) {
                systemId = "file:x-jax-ws-include-" + i++;
                Source src2 = this.createSameTnsPseudoSchema((String)e.getKey(), sameTnsSchemas, systemId);
                mdresolver.addSchema(src2);
            }
            else {
                systemId = (String)sameTnsSchemas.get(0);
            }
            oneSchemaForTns.put(e.getKey(), systemId);
        }
        Source pseudoSchema = this.createMasterPseudoSchema(oneSchemaForTns);
        return new Source[] { pseudoSchema };
    }
    
    @Nullable
    private void addSchemaFragmentSource(Document doc, String systemId, Map<String, DOMSource> map) {
        Element e = doc.getDocumentElement();
        assert !(!e.getNamespaceURI().equals("http://schemas.xmlsoap.org/wsdl/"));
        assert !(!e.getLocalName().equals("definitions"));
        NodeList typesList = e.getElementsByTagNameNS("http://schemas.xmlsoap.org/wsdl/", "types");
        for (int i = 0; i < typesList.getLength(); ++i) {
            NodeList schemaList = ((Element)typesList.item(i)).getElementsByTagNameNS("http://www.w3.org/2001/XMLSchema", "schema");
            for (int j = 0; j < schemaList.getLength(); ++j) {
                Element elem = (Element)schemaList.item(j);
                NamespaceSupport nss = new NamespaceSupport();
                this.buildNamespaceSupport(nss, elem);
                this.patchDOMFragment(nss, elem);
                String docId = systemId + "#schema" + j;
                map.put(docId, new DOMSource(elem, docId));
            }
        }
        return;
    }
    
    private void buildNamespaceSupport(NamespaceSupport nss, Node node) {
        if (node == null || node.getNodeType() != 1) {
            return;
        }
        this.buildNamespaceSupport(nss, node.getParentNode());
        nss.pushContext();
        NamedNodeMap atts = node.getAttributes();
        for (int i = 0; i < atts.getLength(); ++i) {
            Attr a = (Attr)atts.item(i);
            if ("xmlns".equals(a.getPrefix())) {
                nss.declarePrefix(a.getLocalName(), a.getValue());
            }
            else if ("xmlns".equals(a.getName())) {
                nss.declarePrefix("", a.getValue());
            }
        }
        return;
    }
    
    @Nullable
    private void patchDOMFragment(NamespaceSupport nss, Element elem) {
        NamedNodeMap atts = elem.getAttributes();
        Enumeration<String> en = nss.getPrefixes();
        while (en.hasMoreElements()) {
            String prefix = (String)en.nextElement();
            for (int i = 0; i < atts.getLength(); ++i) {
                Attr a = (Attr)atts.item(i);
                if (!"xmlns".equals(a.getPrefix()) || !a.getLocalName().equals(prefix)) {
                    if (AbstractSchemaValidationTube.LOGGER.isLoggable(Level.FINE)) {
                        AbstractSchemaValidationTube.LOGGER.log(Level.FINE, "Patching with xmlns:{0}={1}", new Object[] { prefix, nss.getURI(prefix) });
                    }
                    elem.setAttributeNS("http://www.w3.org/2000/xmlns/", "xmlns:" + prefix, nss.getURI(prefix));
                }
            }
        }
    }
    
    @Nullable
    private Source createSameTnsPseudoSchema(String tns, Collection<String> docs, final String pseudoSystemId) {
        assert docs.size() > 1;
        final StringBuilder sb = new StringBuilder("<xsd:schema xmlns:xsd='http://www.w3.org/2001/XMLSchema'");
        if (!tns.equals("")) {
            sb.append(" targetNamespace='").append(tns).append("'");
        }
        sb.append(">\n");
        for (String systemId : docs) {
            sb.append("<xsd:include schemaLocation='").append(systemId).append("'/>\n");
        }
        sb.append("</xsd:schema>\n");
        if (AbstractSchemaValidationTube.LOGGER.isLoggable(Level.FINE)) {
            AbstractSchemaValidationTube.LOGGER.log(Level.FINE, "Pseudo Schema for the same tns={0}is {1}", new Object[] { tns, sb });
        }
        return new StreamSource(pseudoSystemId) {
            @Override
            public Reader getReader() {
                return new StringReader(sb.toString());
            }
        };
    }
    
    private Source createMasterPseudoSchema(Map<String, String> docs) {
        final StringBuilder sb = new StringBuilder("<xsd:schema xmlns:xsd='http://www.w3.org/2001/XMLSchema' targetNamespace='urn:x-jax-ws-master'>\n");
        for (Map.Entry<String, String> e : docs.entrySet()) {
            String systemId = (String)e.getValue();
            String ns = (String)e.getKey();
            sb.append("<xsd:import schemaLocation='").append(systemId).append("'");
            if (!ns.equals("")) {
                sb.append(" namespace='").append(ns).append("'");
            }
            sb.append("/>\n");
        }
        sb.append("</xsd:schema>");
        if (AbstractSchemaValidationTube.LOGGER.isLoggable(Level.FINE)) {
            AbstractSchemaValidationTube.LOGGER.log(Level.FINE, "Master Pseudo Schema = {0}", sb);
        }
        return new StreamSource("file:x-jax-ws-master-doc") {
            @Override
            public Reader getReader() {
                return new StringReader(sb.toString());
            }
        };
    }
    
    protected void doProcess(Packet packet) throws SAXException {
        this.getValidator().reset();
        Class<? extends ValidationErrorHandler> handlerClass = this.feature.getErrorHandler();
        ValidationErrorHandler handler;
        try {
            handler = (ValidationErrorHandler)handlerClass.newInstance();
        }
        catch (Exception e) {
            throw new WebServiceException(e);
        }
        handler.setPacket(packet);
        this.getValidator().setErrorHandler(handler);
        Message msg = packet.getMessage().copy();
        Source source = msg.readPayloadAsSource();
        try {
            this.getValidator().validate(source);
        }
        catch (IOException e2) {
            throw new WebServiceException(e2);
        }
    }
    
    private String getTargetNamespace(DOMSource src) {
        Element elem = (Element)src.getNode();
        return elem.getAttribute("targetNamespace");
    }
    
    static {
        LOGGER = Logger.getLogger(AbstractSchemaValidationTube.class.getName());
    }
    
    private static class ValidationDocumentAddressResolver implements DocumentAddressResolver
    {
        @Nullable
        public String getRelativeAddressFor(@NotNull SDDocument current, @NotNull SDDocument referenced) {
            AbstractSchemaValidationTube.LOGGER.log(Level.FINE, "Current = {0} resolved relative={1}", new Object[] { current.getURL(), referenced.getURL() });
            return referenced.getURL().toExternalForm();
        }
    }
    
    protected class MetadataResolverImpl implements SDDocumentResolver, LSResourceResolver
    {
        final Map<String, SDDocument> docs;
        final Map<String, SDDocument> nsMapping;
        
        public MetadataResolverImpl() {
            this.docs = new HashMap();
            this.nsMapping = new HashMap();
        }
        
        public MetadataResolverImpl(Iterable<SDDocument> it) {
            this.docs = new HashMap();
            this.nsMapping = new HashMap();
            for (SDDocument doc : it) {
                if (doc.isSchema()) {
                    this.docs.put(doc.getURL().toExternalForm(), doc);
                    this.nsMapping.put(((SDDocument.Schema)doc).getTargetNamespace(), doc);
                }
                else {
                    continue;
                }
            }
        }
        
        void addSchema(Source schema) {
            assert schema.getSystemId() != null;
            String systemId = schema.getSystemId();
            try {
                XMLStreamBufferResult xsbr = (XMLStreamBufferResult)XmlUtil.identityTransform(schema, new XMLStreamBufferResult());
                SDDocumentSource sds = SDDocumentSource.create(new URL(systemId), xsbr.getXMLStreamBuffer());
                SDDocument sdoc = SDDocumentImpl.create(sds, new QName(""), new QName(""));
                this.docs.put(systemId, sdoc);
                this.nsMapping.put(((SDDocument.Schema)sdoc).getTargetNamespace(), sdoc);
            }
            catch (Exception ex) {
                AbstractSchemaValidationTube.LOGGER.log(Level.WARNING, "Exception in adding schemas to resolver", ex);
            }
            return;
        }
        
        void addSchemas(Collection<? extends Source> schemas) {
            for (Source src : schemas) {
                this.addSchema(src);
            }
        }
        
        public SDDocument resolve(String systemId) {
            SDDocument sdi = (SDDocument)this.docs.get(systemId);
            if (sdi == null) {
                SDDocumentSource sds;
                try {
                    sds = SDDocumentSource.create(new URL(systemId));
                }
                catch (MalformedURLException e) {
                    throw new WebServiceException(e);
                }
                sdi = SDDocumentImpl.create(sds, new QName(""), new QName(""));
                this.docs.put(systemId, sdi);
                return sdi;
            }
            return sdi;
        }
        
        public LSInput resolveResource(String type, String namespaceURI, String publicId, String systemId, String baseURI) {
            if (AbstractSchemaValidationTube.LOGGER.isLoggable(Level.FINE)) {
                AbstractSchemaValidationTube.LOGGER.log(Level.FINE, "type={0} namespaceURI={1} publicId={2} systemId={3} baseURI={4}", new Object[] { type, namespaceURI, publicId, systemId, baseURI });
            }
            try {
                SDDocument doc;
                if (systemId == null) {
                    doc = (SDDocument)this.nsMapping.get(namespaceURI);
                }
                else {
                    URI rel = (baseURI != null) ? new URI(baseURI).resolve(systemId) : new URI(systemId);
                    doc = (SDDocument)this.docs.get(rel.toString());
                }
                if (doc != null) {
                    return new LSInput() {
                        @Override
                        public Reader getCharacterStream() {
                            return null;
                        }
                        
                        @Override
                        public void setCharacterStream(Reader characterStream) {
                            throw new UnsupportedOperationException();
                        }
                        
                        @Override
                        public InputStream getByteStream() {
                            ByteArrayBuffer bab = new ByteArrayBuffer();
                            try {
                                doc.writeTo(null, AbstractSchemaValidationTube.this.resolver, bab);
                            }
                            catch (IOException ioe) {
                                throw new WebServiceException(ioe);
                            }
                            return bab.newInputStream();
                        }
                        
                        @Override
                        public void setByteStream(InputStream byteStream) {
                            throw new UnsupportedOperationException();
                        }
                        
                        @Override
                        public String getStringData() {
                            return null;
                        }
                        
                        @Override
                        public void setStringData(String stringData) {
                            throw new UnsupportedOperationException();
                        }
                        
                        @Override
                        public String getSystemId() {
                            return doc.getURL().toExternalForm();
                        }
                        
                        @Override
                        public void setSystemId(String systemId) {
                            throw new UnsupportedOperationException();
                        }
                        
                        @Override
                        public String getPublicId() {
                            return null;
                        }
                        
                        @Override
                        public void setPublicId(String publicId) {
                            throw new UnsupportedOperationException();
                        }
                        
                        @Override
                        public String getBaseURI() {
                            return doc.getURL().toExternalForm();
                        }
                        
                        @Override
                        public void setBaseURI(String baseURI) {
                            throw new UnsupportedOperationException();
                        }
                        
                        @Override
                        public String getEncoding() {
                            return null;
                        }
                        
                        @Override
                        public void setEncoding(String encoding) {
                            throw new UnsupportedOperationException();
                        }
                        
                        @Override
                        public boolean getCertifiedText() {
                            return false;
                        }
                        
                        @Override
                        public void setCertifiedText(boolean certifiedText) {
                            throw new UnsupportedOperationException();
                        }
                    };
                }
            }
            catch (Exception e) {
                AbstractSchemaValidationTube.LOGGER.log(Level.WARNING, "Exception in LSResourceResolver impl", e);
            }
            if (AbstractSchemaValidationTube.LOGGER.isLoggable(Level.FINE)) {
                AbstractSchemaValidationTube.LOGGER.log(Level.FINE, "Don''t know about systemId={0} baseURI={1}", new Object[] { systemId, baseURI });
            }
            return null;
        }
    }
}
